"""Time series of type DICTIONARY should reference an existing dictionary.

<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Values in a time series of type DICTIONARY map to textual labels.
This conveys meaning to analytics users.
A time series that is missing this dictionary will be confusing as the user
will be presented with numerical values only.
</p>
</div>
"""

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Dictionary is present"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Data types",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "data_type": [DataType.DICTIONARY],
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    dictionary = analysis_input.metadata.get_field(fields.Dictionary)
    has_dictionary = dictionary is not None and len(dictionary.mapping) > 0

    return timeseer.AnalysisResult(
        check_results=[
            timeseer.CheckResult(_CHECK_NAME, float(not has_dictionary)),
        ],
    )



""" Calculate the distributed jump sketches of the time series. """

import jsonpickle

from ddsketch.ddsketch import DDSketch
from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType


META: dict = {
    "statistics": [
        {"name": "Jump Up Sketch"},
        {"name": "Jump Down Sketch"},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64, None],
        }
    ],
    "signature": "univariate",
}


def _clean_dataframe(df):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_get_sketches(analysis_input, up_sketch, down_sketch):
    df = _clean_dataframe(analysis_input.data)
    if len(df) == 0:
        return up_sketch, down_sketch

    if up_sketch is None:
        up_sketch = DDSketch(0.001)
    if down_sketch is None:
        down_sketch = DDSketch(0.001)

    value_diff = df["value"].diff()

    up_values = value_diff[value_diff > 0]
    down_values = value_diff[value_diff < 0]

    for v in up_values:
        up_sketch.add(v)

    for v in down_values:
        down_sketch.add(v)

    return up_sketch, down_sketch


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return jsonpickle.decode(statistics[0])


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> bool:
    if is_string_dtype(analysis_input.data["value"]):
        return False
    return True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if not _is_valid_input(analysis_input):
        return timeseer.AnalysisResult()

    previous_up_sketch = _get_relevant_statistic(analysis_input, "Jump Up Sketch")
    previous_down_sketch = _get_relevant_statistic(analysis_input, "Jump Down Sketch")

    up_sketch, down_sketch = _run_get_sketches(
        analysis_input, previous_up_sketch, previous_down_sketch
    )

    statistics = []
    if up_sketch is not None:
        statistics.append(
            timeseer.Statistic(
                META["statistics"][0]["name"], "sketch", jsonpickle.encode(up_sketch)
            )
        )
    if down_sketch is not None:
        statistics.append(
            timeseer.Statistic(
                META["statistics"][1]["name"], "sketch", jsonpickle.encode(down_sketch)
            )
        )

    if len(statistics) == 0:
        return timeseer.AnalysisResult()

    return timeseer.AnalysisResult(statistics=statistics)



"""The data in a time series should be present within the analysis timeframe.

<p>
The data in a time series should be present within the analysis timeframe.
</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
No input - No output.
</p>
</div>
"""

import timeseer

_CHECK_NAME = "Data available"

META = {
    "checks": [{"name": _CHECK_NAME, "kpi": "Completeness", "data_type": "bool"}],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    no_data = len(analysis_input.data) == 0
    check = timeseer.CheckResult(_CHECK_NAME, float(no_data))
    return timeseer.AnalysisResult([check])



"""There should be no consecutively recorded repeated values close to each other.

<p>This is a simple check for undercompression. If there are repeated
consecutive values recorded closer together than the maximum allowed time frame
by the compression settings, this could indicate that the exception setting is too strict.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Undercompression can lead to too high storage requirements.
</p>
</div>
"""

from typing import List

import jsonpickle
import numpy as np
import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
    calculate_local_score,
)

_CHECK_NAME = "No repeated values"
_EVENT_FRAME_NAME = "Repeated values"


META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Consistency",
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
    "signature": "univariate",
}


def _get_active_points(df, cut_off):
    short_intervals = (df.index.to_series().diff()).dt.total_seconds() < cut_off
    same_value = df["value"].diff() == 0
    return short_intervals & same_value


def _get_intervals(active_points, df, event_type):
    interval_grp = (active_points != active_points.shift().bfill()).cumsum()
    active_points[active_points.isna()] = 0
    active_points = np.array(active_points, dtype=bool)
    intervals = (
        df.assign(interval_grp=interval_grp)[active_points]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _are_consecutive_duplicates_present(
    analysis_input, median_archival_step, open_event_frames, stale_sketch
) -> int:
    df = _clean_dataframe(analysis_input.data)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    cut_off = stale_sketch.get_quantile_value(0.1)

    active_points = _get_active_points(df, cut_off)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = _handle_open_intervals(df, intervals)
    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )
    percentage = calculate_local_score(
        df, analysis_start, intervals, median_archival_step
    )

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-1]

    return percentage, list(frames), last_analyzed_point


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


# pylint: disable=missing-function-docstring
def run(analysis_input: timeseer.AnalysisInput) -> timeseer.AnalysisResult:
    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    json_stale_sketch = _get_relevant_statistic(analysis_input, "Archival Sketch")
    if median_archival_step is None or json_stale_sketch is None:
        return timeseer.AnalysisResult()
    stale_sketch = jsonpickle.decode(json_stale_sketch)

    open_event_frames = _get_open_event_frames(analysis_input)

    pct_active, frames, last_analyzed_point = _are_consecutive_duplicates_present(
        analysis_input, median_archival_step, open_event_frames, stale_sketch
    )

    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult(
        [check], frames, last_analyzed_point=last_analyzed_point
    )



"""check for a change after a critical event based on Kolmogorov-Smirnoff.

<p>This check identifies whether there is a significant difference in the
distribution of a series before and after a critical event.</p>
<div class="ts-check-impact">
<p>Changes in the behavior of a series after a critical event, such as a
maintenance, turnaround or other types of interventions could indicate
human error. This can range from physical damage to a sensor, to forgotten
control strategy updates.</p>
</div>"""

import pandas as pd

from scipy import stats

from timeseer import (
    AnalysisResult,
    CheckResult,
    ComparisonAnalysisInput,
    DataType,
    ProcessType,
)


_CHECK_NAME = "Comparison change ks"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
            "process_type": [ProcessType.CONTINUOUS, ProcessType.REGIME],
        }
    ],
    "signature": "comparison",
}


def _clean_data(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_compare_ks(analysis_input: ComparisonAnalysisInput):
    assert analysis_input.reference_data is not None
    df_before = _clean_data(analysis_input.reference_data["value"])
    df_after = _clean_data(analysis_input.data["value"])

    if len(df_before) == 0 or len(df_after) == 0:
        return None

    _, p_value = stats.ttest_ind(df_before, df_after)
    return p_value >= 0.05


def run(
    analysis_input: ComparisonAnalysisInput,
):  # pylint: disable=missing-function-docstring
    is_change = _run_compare_ks(analysis_input)
    if is_change is None:
        return AnalysisResult()

    check = CheckResult(_CHECK_NAME, float(is_change))

    return AnalysisResult(check_results=[check])



"""Min, mean and max time between archival.

<p>In historians points are often stored not at fixed time intervals,
but based on approximations. This means that the rate at which points are stored
(archived) is not fixed.
These statistics also provide initial insights in certain data captation issues.
For example a 0 or negative minimal archival points to timestamp duplicates or out of order samples.</p>"""

import pandas as pd
import numpy as np

import timeseer


META: dict = {
    "statistics": [
        {"name": "Archival time mean"},
        {"name": "Archival time min"},
        {"name": "Archival time max"},
        {"name": "Archival time median"},
    ],
    "run": "before",
    "conditions": [{"min_series": 1, "min_weeks": 1, "min_data_points": 300}],
    "signature": "univariate",
}


def _run_ts_archival_timing(df: pd.DataFrame):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return (
        np.min(diff_times).total_seconds(),
        np.mean(diff_times).total_seconds(),
        np.max(diff_times).total_seconds(),
        pd.Timedelta(np.median(diff_times)).total_seconds(),
    )


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:

    (
        archival_min,
        archival_mean,
        archival_max,
        archival_median,
    ) = _run_ts_archival_timing(analysis_input.data)
    check_mean = timeseer.Statistic(
        META["statistics"][0]["name"], "hidden", float(archival_mean)
    )
    check_min = timeseer.Statistic(
        META["statistics"][1]["name"], "hidden", float(archival_min)
    )
    check_max = timeseer.Statistic(
        META["statistics"][2]["name"], "hidden", float(archival_max)
    )
    check_median = timeseer.Statistic(
        META["statistics"][3]["name"], "hidden", float(archival_median)
    )

    values = [
        ("Min", float(archival_min)),
        ("Max", float(archival_max)),
        ("Mean", float(archival_mean)),
        ("Median", float(archival_median)),
    ]
    table_statistics = timeseer.Statistic("Archival statistics", "table", values)
    return timeseer.AnalysisResult(
        statistics=[check_min, check_mean, check_max, check_median, table_statistics]
    )



"""Data type deduced by actual values.

<p>This check tries to identify the type of data that is being measured:
<ul>
<li>FLOAT32/FLOAT64</li>
<li>CATEGORICAL</li>
<li>STRING</li>
<li>DICTIONARY</li>
</ul>
</p>
<div class="ts-check-impact">
<p>Not having the right data type can cause issues with interpretation of the data and
automatization of analytics solutions. For instance if a series is misinterpreted as
FLOAT64 but in reality is a categorical values then performing calculations on this serie
might lead to unexpected result. In categorical series it is not necessart that labels
relate to distance. E.g. labels (0, ..., 10) does not mean that values of 0 are further
away from 10 than values of 9. They are merely different categories.
</p>
</div>
"""

from pandas.api.types import is_string_dtype

import numpy as np
import timeseer


META = {
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


# Code is duplicated in deduced_process_type
def _check_data_type(df):  # pylint: disable=too-many-return-statements
    if is_string_dtype(df["value"]):
        return "STRING"
    if (
        df["value"].nunique() <= 10 and all(isinstance(x, int) for x in df["value"])
    ) or all(x in range(101) for x in df["value"].to_numpy()):
        return "CATEGORICAL"
    if isinstance(df["value"][0], np.float64):
        return "FLOAT64"
    if isinstance(df["value"][0], np.float32):
        return "FLOAT32"
    if isinstance(df["value"][0], (np.int, np.int16, np.int64, np.int32)):
        return "FLOAT32"
    return None


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    result = _check_data_type(analysis_input.data)
    if result is None:
        return timeseer.AnalysisResult()

    calculated_metadata = [timeseer.CalculatedMetadata("data type", result)]
    return timeseer.AnalysisResult(calculated_metadata=calculated_metadata)



"""Utility functions for unit conversions."""

from pathlib import Path
from pint import UnitRegistry


def get_unit_registry() -> UnitRegistry:
    """Return a pint UnitRegistry configured for Timeseer."""
    return UnitRegistry(str(Path(__file__).parent.resolve()) + "/default_replace.txt")



import sqlite3

from yoyo import step

__depends__ = {"052__update_flow_evaluation_groups"}


def add_foreign_keys(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute("pragma schema_version")
    (schema_version,) = cursor.fetchone()
    cursor.execute("pragma writable_schema = ON")
    cursor.execute(
        """
        update sqlite_master
        set sql = 'CREATE TABLE FlowEvaluations (
    id integer primary key autoincrement,
    group_id integer not null,
    series_set_name text not null,
    foreign key (group_id) references FlowEvaluationGroups(id) on delete cascade
)'
        where type = 'table' and name = 'FlowEvaluations'
    """
    )
    cursor.execute(f"pragma schema_version = {schema_version + 1}")
    cursor.execute("pragma writable_schema = OFF")


steps = [
    step(add_foreign_keys),
]



""" Calculate the min, mean ad max of an (unevenly spaced) time series. """

import numpy as np
import pandas as pd

from pandas.api.types import is_string_dtype
from statsmodels.stats.weightstats import DescrStatsW

import timeseer

from timeseer import DataType
from timeseer.metadata import fields


META: dict = {
    "statistics": [
        {"name": "Mean"},
        {"name": "Min"},
        {"name": "Max"},
        {"name": "Median"},
        {"name": "Standard Deviation"},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


def _get_weights_unequally_spaced(df: pd.DataFrame):
    times = (
        (df.index.to_series() - df.index.to_series().shift()).dt.total_seconds()
    ).to_numpy()
    total_duration = np.nansum(times)
    weights = np.empty(len(times))
    weights[0] = 0.5 * times[1] / total_duration
    weights[-1] = 0.5 * times[-1] / total_duration
    if len(times) > 2:
        for i in range(1, (len(times) - 1)):
            weights[i] = 0.5 * (times[i] + times[i + 1]) / total_duration
    return weights


def _run_ts_mean(values, weights):
    return DescrStatsW(values, weights=weights).mean


def _run_ts_std(values, weights):
    return DescrStatsW(values, weights=weights).std


def _run_ts_median(values, weights):
    return DescrStatsW(values, weights).quantile(np.array([0.5]))


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> bool:
    if is_string_dtype(analysis_input.data["value"]):
        return False
    if len(_clean_dataframe(analysis_input.data).dropna()) < 2:
        return False

    data_type = analysis_input.metadata.get_field(fields.DataType)
    if data_type not in [DataType.FLOAT64, DataType.FLOAT32, None]:
        return False
    return True


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if not _is_valid_input(analysis_input):
        return timeseer.AnalysisResult()

    clean_df = _clean_dataframe(analysis_input.data)
    values = clean_df["value"].to_numpy()
    weights = _get_weights_unequally_spaced(clean_df)

    weights = weights[~np.isnan(values)]
    values = values[~np.isnan(values)]

    mean = _run_ts_mean(values, weights)
    median = _run_ts_median(values, weights)
    std = _run_ts_std(values, weights)
    stat_mean = timeseer.Statistic(META["statistics"][0]["name"], "hidden", float(mean))
    stat_min = timeseer.Statistic(
        META["statistics"][1]["name"],
        "hidden",
        float(analysis_input.data["value"].min(skipna=True)),
    )
    stat_max = timeseer.Statistic(
        META["statistics"][2]["name"],
        "hidden",
        float(analysis_input.data["value"].max(skipna=True)),
    )
    stat_median = timeseer.Statistic(
        META["statistics"][3]["name"], "hidden", float(median)
    )
    stat_std = timeseer.Statistic(META["statistics"][4]["name"], "hidden", float(std))

    values = [
        ("Min", float(analysis_input.data["value"].min(skipna=True))),
        ("Max", float(analysis_input.data["value"].max(skipna=True))),
        ("Mean", float(mean)),
        ("median", float(median)),
        ("Std", float(std)),
    ]
    table_statistics = timeseer.Statistic("Value statistics", "table", values)
    statistics = [
        stat_mean,
        stat_min,
        stat_max,
        stat_median,
        stat_std,
        table_statistics,
    ]
    return timeseer.AnalysisResult(
        statistics=statistics,
    )



"""Identify series based on deduced regex.

<p>This returns suggestions for pairs to be treated as series group.</p>"""

from typing import List
from uuid import UUID

import numpy as np
import pandas as pd

import timeseer

from timeseer import AnalysisInput, AnalysisResult, SeriesId


_CHECK_NAME = "Series set regex discovery"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 3,
            "min_weeks": 0,
            "min_data_points": 0,
        }
    ],
}


def _determine_char_type(char):
    if char.isnumeric():
        return "DIGIT"
    if char.isalpha():
        return "CHARACTER"
    if char.isspace():
        return "WHITESPACE"
    return "SPECIAL"


def _get_current_regex_part(current_type, name, last_index_of_type, idx):
    if current_type == "DIGIT":
        regex_part = "\\d{" + str(last_index_of_type) + "}"
    elif current_type == "CHARACTER":
        regex_part = "[a-zA-Z]{" + str(last_index_of_type) + "}"
    elif current_type == "WHITESPACE":
        regex_part = "\\s{" + str(last_index_of_type) + "}"
    elif current_type == "SPECIAL":
        regex_part = "[" + name[idx - 1] + "]"
    return regex_part


def _create_generic_regex(name, idx, last_index_of_type, regex_part):
    if (idx - last_index_of_type) <= 0:
        return "^" + regex_part
    if (idx + last_index_of_type) > len(name):
        return ".*" + regex_part + "$"
    return ".{" + str(idx - last_index_of_type) + "}" + regex_part


def _get_current_regex_part_for_last_char(name, char, idx):
    if _determine_char_type(char) == "DIGIT":
        regex_part = "\\d{1}"
    elif _determine_char_type(char) == "CHARACTER":
        regex_part = "[a-zA-Z]{1}"
    elif _determine_char_type(char) == "WHITESPACE":
        regex_part = "\\s{1}"
    elif _determine_char_type(char) == "SPECIAL":
        regex_part = "[" + name[idx] + "]"
    return regex_part


def _get_regex_based_on_name(name):
    last_index_of_type = -1
    current_type = _determine_char_type(name[0])
    regex_string = ""
    generic_regex_for_name = []

    for idx, char in enumerate(name):
        last_index_of_type += 1

        # Check whether we have to close a logical part
        if (
            _determine_char_type(char) != current_type
            or _determine_char_type(char) == "SPECIAL"
            or idx == len(name) - 1
        ):

            last_char_different = False

            # If last char has same type, extend length of current type
            if idx == len(name) - 1:
                if _determine_char_type(char) == current_type:
                    last_index_of_type += 1
                if _determine_char_type(char) != current_type:
                    last_char_different = True

            regex_part = _get_current_regex_part(
                current_type, name, last_index_of_type, idx
            )

            regex_string += regex_part
            if current_type != "SPECIAL":
                generic_regex_for_name.append(
                    _create_generic_regex(name, idx, last_index_of_type, regex_part)
                )

            if last_char_different:
                regex_part = _get_current_regex_part_for_last_char(name, char, idx)
                regex_string += regex_part
                generic_regex_for_name.append(
                    _create_generic_regex(name, idx, last_index_of_type, regex_part)
                )

            current_type = _determine_char_type(char)
            last_index_of_type = 0
    return regex_string, generic_regex_for_name


def _get_series_regex_matrix(names):
    """This function finds:
     * all global regex - i.e. trying to match entire series names and
     * generic regex subsets of these global regex containing sequences of [a-zA-Z] or digits
    Several extensions are possible:
     * creating specific regex by taking parts of the actual series names
     * grouping based on np.unique on location wise matches to the generic regex subsets
     * removing location specific matches
     * removing start / end matches
    """
    series_regex_matrix = pd.DataFrame(data={"blank": False}, index=names)

    generic_regex = set()

    # Global complete RegEx structure
    while not series_regex_matrix.any(1).all():
        current_name = series_regex_matrix[(~series_regex_matrix).all(1)].index[0]

        regex_string, generic_for_name = _get_regex_based_on_name(current_name)

        matches = series_regex_matrix.index.str.match(regex_string)
        matches_column = pd.DataFrame(data={regex_string: matches}, index=names)
        series_regex_matrix = pd.concat([series_regex_matrix, matches_column], axis=1)

        generic_regex = generic_regex.union(generic_for_name)

    # For each generic RegEx add a column
    for expression in generic_regex:
        matches = series_regex_matrix.index.str.match(expression)
        matches_column = pd.DataFrame(data={expression: matches}, index=names)
        series_regex_matrix = pd.concat([series_regex_matrix, matches_column], axis=1)
    return series_regex_matrix


def _run_regex_recommendation(inputs: List[AnalysisInput]) -> List[List[UUID]]:
    relevant_series_ids = np.array(
        [analysis_input.metadata.series.series_id for analysis_input in inputs]
    )
    names = np.array(
        [
            analysis_input.metadata.series.name
            for analysis_input in inputs
            if analysis_input.metadata.series.name not in ("", None)
        ]
    )

    # Create matrix and only keep non-trivial groups (2-4) members for now
    min_members = 2
    max_members = len(inputs) - 1
    series_regex_matrix = _get_series_regex_matrix(names).reset_index(drop=True)
    relevant_idx = (series_regex_matrix.sum() >= min_members) & (
        series_regex_matrix.sum() <= max_members
    )
    rel_groups = series_regex_matrix.loc[:, relevant_idx]

    connected_groups = []
    for i in range(rel_groups.shape[1]):
        connected_groups.append(
            np.array(rel_groups.reset_index(drop=True).index[rel_groups.iloc[:, i]])
        )

    unique_cg = [list(x) for x in set(tuple(x) for x in connected_groups)]

    return [relevant_series_ids[group] for group in unique_cg]


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = analysis_input.inputs

    suggestions = _run_regex_recommendation(inputs)
    return AnalysisResult(
        series_sets={
            "naming pattern": [
                [SeriesId(suggestion) for suggestion in group] for group in suggestions
            ],
        }
    )



"""The unit should not be one non-letter character or all the same non-letter characters.

<p>Units aid in the interpretation of measurement readings. Often series naming conventions
show some insight into the type of sensor (e.g. PI for pressure indicator).
Due to time constraints or as a placeholder often "--" or similar is used.</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check belongs to the score group Unit.
If this check is True, it means that the score for group Unit will be 0%.</p>
<div class="ts-check-impact">
<p>
Interpreting a dimensionless variable can lead to wrong conclusions and actions
(e.g. bara vs barg vs kPa vs ...). When combining data from different sensors of a similar type,
comparisons without units are impossible. Without units the interpretation of the sensibility of
a measurement value is hindered.
</p>
</div>
"""

import timeseer

from timeseer.metadata import fields

_CHECK_NAME = "Unit is interpretable"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Unit",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)
    if unit == "":
        return timeseer.AnalysisResult()
    unique_chars = set(unit)
    different_character_count = len(unique_chars)
    score = different_character_count == 1 and unit[0] in [">", "-", "<"]
    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, float(score))]
    )



"""Timeseer Client allows querying of data and metadata."""

import json
import time

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Union

from kukur import Metadata, SeriesSelector
from kukur.client import Client as KukurClient

import pyarrow as pa
import pyarrow.flight as fl


class Client(KukurClient):
    """Client connects to Timeseer using Arrow Flight."""

    def upload_data(
        self,
        metadata_or_data: Union[Metadata, list[tuple[Metadata, pa.Table]]],
        table: Optional[pa.Table] = None,
        *,
        analyze=True,
        block=True
    ):
        """Upload time series data to Timeseer.

        This requires a configured 'flight-upload' source in Timeseer.

        There are two ways to call this method.

        One requires two arguments:
            metadata: any known metadata about the time series. This will be merged with the metadata
                already known by Timeseer depending on the source configuration. The source of the series should match
                the source name of a 'flight-upload' source.
            table: a pyarrow.Table of two columns.
                The first column with name 'ts' contains Arrow timestamps.
                The second column with name 'value' contains the values as a number or string.

        The second accepts a list of tuples of the same arguments. This allows uploading multiple time series at the
        same time.

        When `analyze` is `True`, start a flow evaluation.
        When `block` is `True`, block execution until the flow evaluation is done.
        """
        if table is not None:
            assert isinstance(metadata_or_data, Metadata)
            self._upload_data_single(metadata_or_data, table, analyze, block)
        else:
            assert not isinstance(metadata_or_data, Metadata)
            self._upload_data_multiple(metadata_or_data, analyze, block)

    def _upload_data_single(
        self, metadata: Metadata, table: pa.Table, analyze: bool, block: bool
    ):
        self._upload_data_multiple([(metadata, table)], analyze, block)

    def _upload_data_multiple(
        self, many_series: list[tuple[Metadata, pa.Table]], analyze: bool, block: bool
    ):
        client = self._get_client()
        selectors = []
        for metadata, table in many_series:
            metadata_json = metadata.to_data()
            selector = SeriesSelector(metadata.series.source, metadata.series.name)
            selectors.append(selector)
            descriptor = fl.FlightDescriptor.for_command(
                json.dumps(
                    dict(
                        metadata=metadata_json,
                    )
                )
            )
            writer, reader = client.do_put(descriptor, table.schema)
            writer.write_table(table)
            writer.done_writing()
            buf: pa.Buffer = reader.read()
            response: dict = json.loads(buf.to_pybytes())
            writer.close()

        if "flowName" in response:
            if analyze:
                _analyze_flow(
                    client, response["flowName"], limitations=selectors, block=block
                )
        elif block:
            _wait_for_flow_evaluation(client, response)

    def evaluate_flow(self, flow_name: str, *, block=True):
        """Evaluate a flow.

        Args:
            flow_name: the name of the flow to evaluate
            block: block until the evaluation completes (keyword-only, default True)
        """
        client = self._get_client()
        _analyze_flow(client, flow_name, block=block)

    def get_event_frames(
        self,
        selector: SeriesSelector,
        start_date: datetime = None,
        end_date: datetime = None,
        frame_type: Union[str, list[str]] = None,
    ) -> pa.Table:
        """Get all event frames matching the given criteria.

        Args:
            selector: the time series source, exposed flow or time series to which the event frames are linked.
            start_date: the start date of the range to find overlapping event frames in. Defaults to one year ago.
            end_date: the end date of the range to find overlapping event frames in. Defaults to now.
            frame_type: the type or types of event frames to search for. Finds all types when empty.

        Returns::
            A pyarrow Table with 4 columns.
            The first column ('start_date') contains the start date.
            The second column ('end_date') contains the end date.
            The third column ('type') contains the type of the returned event frame as a string.
            Columns 4 contains possible multiple references for the event frame.
        """
        if start_date is None or end_date is None:
            now = datetime.utcnow().replace(tzinfo=timezone(timedelta(0)))
            if start_date is None:
                start_date = now.replace(year=now.year - 1)
            if end_date is None:
                end_date = now

        query: Dict[str, Any] = {
            "query": "get_event_frames",
            "selector": {"source": selector.source},
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }

        if selector.name is not None:
            query["selector"]["name"] = selector.name

        if frame_type is not None:
            query["type"] = frame_type

        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def get_data_quality_score_data_sources(
        self, source_names: List[str]
    ) -> dict[str, int]:
        """Get the data quality score of a data source or exposed flow.

        Args:
            source_names: A list of time series sources or Flight Expose blocks

        Returns::
            A data quality score of every given source as a percentage.
        """
        body = {
            "source_names": source_names,
        }
        results = list(
            self._get_client().do_action(
                ("get_data_quality_score_data_sources", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def get_kpi_scores(
        self,
        source_name: str,
    ) -> dict[str, dict[str, int]]:
        """Get the kpi scores of a data source or exposed flow.

        Args:
            source_name: The time series source or the name of the Flight Expose block.

        Returns::
            The score per KPI as a percentage, keyed by the source_name provided.
        """
        body = {
            "source_name": source_name,
        }

        results = list(
            self._get_client().do_action(("get_kpi_scores", json.dumps(body).encode()))
        )
        return json.loads(results[0].body.to_pybytes())

    def get_series_results(
        self, selector: SeriesSelector
    ) -> list[dict[str, Union[str, float]]]:
        """Return the univariate check results for the given series.

        Args:
            selector: The time series to return results for.

        Returns::
            A list of check results, where each check result is a dictionary with 'name' and 'result'.
            The result is returned as a floating point number between 0.0 (bad) and 1.0 (good)."""
        body = {
            "selector": {
                "source": selector.source,
                "name": selector.name,
            },
        }
        results = list(
            self._get_client().do_action(
                ("get_series_results", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())

    def get_series_statistics(self, selector: SeriesSelector) -> list[dict[str, Any]]:
        """Return the univariate statistics for the given series.

        Args:
            selector: The time series to return statistics for.

        Returns::
            A list of statistics. Each statistic is a dict with three fields:
              - name (str)
              - dataType (str) - one of 'float', 'pct', 'datetime', 'hidden', 'table'
              - result (exact type depending on the dataType)"""
        body = {
            "selector": {
                "source": selector.source,
                "name": selector.name,
            },
        }
        results = list(
            self._get_client().do_action(
                ("get_series_statistics", json.dumps(body).encode())
            )
        )
        return json.loads(results[0].body.to_pybytes())


def _analyze_flow(
    client: fl.FlightClient,
    flow_name: str,
    *,
    limitations: Optional[list[SeriesSelector]] = None,
    block=True
):
    flow: dict = dict(flowName=flow_name)
    if limitations is not None:
        flow["limitations"] = [
            dict(source=selector.source, name=selector.name) for selector in limitations
        ]
    results = list(client.do_action(("evaluate_flow", json.dumps(flow).encode())))
    response = json.loads(results[0].body.to_pybytes())

    if block:
        _wait_for_flow_evaluation(client, response)


def _wait_for_flow_evaluation(client: fl.FlightClient, response: dict):
    while True:
        results = list(
            client.do_action(
                ("get_flow_evaluation_state", json.dumps(response).encode())
            )
        )
        state = json.loads(results[0].body.to_pybytes())
        if (
            state["completed"] == state["total"]
            and state["blockCompleted"] == state["blockTotal"]
        ):
            break
        time.sleep(1)



"""The provided limits are expected to be close to the recommended limits based on historical data.

<p>Limits can be set either in historian databases, manual, via spec sheets etc.
These limits should be related to the expected operating range of the corresponding sensor.
If historical analysis demonstrates that the actual operating range is significantly different
from the provided limits, this could mean that the limits were wrongly initiated.
</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check is True if there is more than 50% difference in any limit or range.</p>
<div class="ts-check-impact">
<p>Conservatively configured limits can lead to a potential alarm flood. Reversily,
too relaxed limits causes missing important warning signals by for instance masking drift.</p>
</div>
"""

import pandas as pd
import numpy as np

from pint import UndefinedUnitError

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import get_unit_registry
from timeseer.metadata import fields


_CHECK_NAME = "Improbable limits"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Improbable limits",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 12,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_dimensionality(unit):
    ureg = get_unit_registry()
    try:
        return ureg.get_dimensionality(unit)
    except (UndefinedUnitError, AttributeError, ValueError, KeyError):
        return "Unknown"


def _make_units_dir():
    ureg = get_unit_registry()
    units_zero_min = [
        "meter",
        "second",
        "ampere",
        "candela",
        "gram",
        "mole",
        "kelvin",
        "unit",
        "pH",
        "m2",
        "liter",
        "hertz",
        "kph",
        "galileo",
        "newton",
        "joule",
        "watt",
        "water",
        "pascal",
        "foot_pound",
        "poise",
        "stokes",
        "rhe",
        "particle",
        "molar",
        "katal",
        "clausius",
        "entropy_unit",
        "curie",
        "langley",
        "nit",
        "lumen",
        "lux",
        "a_u_intensity",
        "volt",
        "ohm",
        "siemens",
        "henry",
        "weber",
        "tesla",
        "bohr_magneton",
    ]
    dimensions_zero_min = []
    for unit in units_zero_min:
        dimension = ureg.get_dimensionality(unit)
        dimensions_zero_min.append(dimension)
    return dimensions_zero_min


def _unit_dir(metadata: timeseer.Metadata):
    input_dimension = _get_dimensionality(metadata.get_field(fields.Unit))
    return input_dimension


def _deconstruct(value):
    if value == 0:
        return [0, 0, 0, 0]
    abs_value = abs(value)
    sign = np.sign(value)
    magnitude = np.floor(np.log10(abs_value))
    scale = np.floor(abs_value // np.power(10, magnitude))
    remainder = abs_value - scale * np.power(10, magnitude)
    smsr = [sign, magnitude, scale, remainder]
    return smsr


def _max_limit(max_value):
    smsr = _deconstruct(max_value)
    sign = smsr[0]
    magnitude = smsr[1]
    scale = smsr[2]
    remainder = smsr[3]
    if magnitude < 0:
        if sign < 0:
            return 0
        lim = np.power(10, magnitude + 1)
        return lim
    lim = np.power(10, magnitude) * (scale)
    if remainder > 0:
        if sign > 0:
            smsr_remainder = _deconstruct(remainder)
            small_mag = max((smsr_remainder[1] + 1), np.floor(magnitude / 2))
            small = np.power(10, np.sign(magnitude) * small_mag)
            lim += small
        if sign < 0:
            smsr_remainder = _deconstruct(remainder)
            if smsr_remainder[1] >= (magnitude / 2):
                small = np.power(10, smsr_remainder[1]) * smsr_remainder[2]
                lim += small
    return float(sign * lim)


def _min_limit(min_value):
    smsr = _deconstruct(min_value)
    sign = smsr[0]
    magnitude = smsr[1]
    scale = smsr[2]
    remainder = smsr[3]
    lim = 0
    if sign < 0:
        if magnitude < 0:
            return -1
        if magnitude >= 0:
            lim = np.power(10, magnitude) * (scale)
            if remainder > 0:
                smsr_remainder = _deconstruct(remainder)
                small_mag = max((smsr_remainder[1] + 1), np.floor(magnitude / 2))
                small = np.power(10, np.sign(magnitude) * small_mag)
                lim += small
    return float(sign * lim)


def _find_recommended_upper_limit(df: pd.DataFrame):
    if len(df["value"]) < 100:
        max_value = max(df["value"])
    else:
        q_1, q_3 = np.quantile(df["value"], [0.25, 0.75])
        iqr = q_3 - q_1
        max_value = max(df["value"][df["value"] <= q_3 + 1.5 * iqr])

    rec_high = 0
    if max_value != 0:
        rec_high = _max_limit(max_value)
    return rec_high


def _find_recommended_lower_limit(df: pd.DataFrame):
    if len(df["value"]) < 100:
        min_value = min(df["value"])
    else:
        q_1, q_3 = np.quantile(df["value"], [0.25, 0.75])
        iqr = q_3 - q_1
        min_value = min(df["value"][df["value"] >= q_1 - 1.5 * iqr])

    rec_low = 0
    if min_value != 0:
        rec_low = _min_limit(min_value)
    return rec_low


def _are_hist_limits_close_to_metadata(
    rec_lower, rec_upper, meta_lower, meta_upper
):  # pylint: disable=too-many-return-statements
    rec_range = rec_upper - rec_lower
    meta_range = meta_upper - meta_lower

    if rec_range == 0 or meta_range == 0:
        return False

    if meta_lower == 0:
        if abs(rec_upper / meta_upper) > 2:
            return False
        if rec_lower <= -1:
            return False
        if rec_upper == 0:
            return False
        if abs(meta_upper / rec_upper) > 2:
            return False
    else:
        if (rec_range / meta_range) > 2:
            return False
        if (meta_range / rec_range) > 2:
            return False
        if abs(rec_upper - meta_upper) > (min(rec_range, meta_range) * 0.5):
            return False
        if abs(rec_lower - meta_lower) > (min(rec_range, meta_range) * 0.5):
            return False
    return True


def _are_limits_probable(metadata: timeseer.Metadata, data: pd.DataFrame):

    dimensions_zero_min = _make_units_dir()

    if _unit_dir(metadata) in dimensions_zero_min:
        rec_lower = 0
    else:
        rec_lower = _find_recommended_lower_limit(data)

    if _unit_dir(metadata) == _get_dimensionality("pH"):
        rec_upper = 14
    else:
        rec_upper = _find_recommended_upper_limit(data)

    if _are_hist_limits_close_to_metadata(
        rec_lower,
        rec_upper,
        metadata.get_field(fields.LimitLowFunctional),
        metadata.get_field(fields.LimitHighFunctional),
    ):
        return 0.0

    return 1.0


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    metadata = analysis_input.metadata
    limit_low = metadata.get_field(fields.LimitLowFunctional)
    limit_high = metadata.get_field(fields.LimitHighFunctional)

    if limit_low is None or limit_high is None:
        return timeseer.AnalysisResult()

    if limit_low >= limit_high:
        return timeseer.AnalysisResult()

    data = analysis_input.data.dropna()
    if len(data) == 0:
        return timeseer.AnalysisResult()

    score = _are_limits_probable(metadata, data)

    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, score)]
    )



"""The data in a time series should match the data type indicated in the metadata.

<p>
Time series data can be numerical or textual.
Operations that have a defined result on numerical data often don't on textual data.
For example, when treated as textual data, 8 is a larger number than 10.
</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Mismatched data types can hamper downstream (automated) analytics.
</p>
</div>
"""

import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Data matches data type"

META = {
    "checks": [{"name": _CHECK_NAME, "kpi": "Integrity", "data_type": "bool"}],
    "conditions": [{"min_series": 1, "min_data_points": 1}],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    data_type = analysis_input.metadata.get_field(fields.DataType)
    if data_type is None:
        return timeseer.AnalysisResult()

    if len(analysis_input.data) == 0:
        is_valid = True
    elif data_type in [DataType.FLOAT32, DataType.FLOAT64, DataType.DICTIONARY]:
        is_valid = pd.api.types.is_numeric_dtype(analysis_input.data["value"])
    elif data_type == DataType.STRING:
        is_valid = pd.api.types.is_string_dtype(analysis_input.data["value"])
    else:
        return timeseer.AnalysisResult()

    return timeseer.AnalysisResult(
        check_results=[
            timeseer.CheckResult(_CHECK_NAME, float(not is_valid)),
        ],
    )



"""Timeseer score module to generate scores for metadata checks.

Scores for metadata checks are calculated based on their metadata group.
If one score indicates a failure, the total score will be a failure.
"""

from collections import defaultdict
from typing import List

from timeseer import CheckResult, ScoreInput, ScoreOutput, ScoreResult

META = {
    "scores": [
        dict(
            name="description",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="description"),
            ],
        ),
        dict(
            name="interpolation type",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="interpolation type"),
            ],
        ),
        dict(
            name="limits",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="limits"),
            ],
        ),
        dict(
            name="unit",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="unit"),
            ],
        ),
        dict(
            name="data types",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="data types"),
            ],
        ),
        dict(
            name="accuracy",
            kpi="Metadata",
            data_type="bool",
            checks=[
                dict(group="accuracy"),
            ],
        ),
    ],
}


def _calculate_score(results: List[CheckResult]) -> float:
    return max([result.result for result in results])


# pylint: disable=missing-function-docstring
def score(
    score_input: ScoreInput,
) -> ScoreOutput:
    metadata_lookup = {meta.name: meta for meta in score_input.check_metadata}

    by_group = defaultdict(list)
    for result in score_input.check_results:
        metadata = metadata_lookup[result.name]
        if metadata.group is None:
            continue
        by_group[metadata.group].append(result)
    scores = []
    for group_name, group_results in by_group.items():
        scores.append(ScoreResult(group_name, _calculate_score(group_results)))
    return ScoreOutput(scores)



"""Data type deduced by actual values.

<p>This check tries to identify the type of process behavior a series abides to:
<ul>
<li>BATCH</li>
<li>GRADE</li>
<li>CONTINUOUS</li>
<li>COUNTER</li>
</ul>
</p>
<div class="ts-check-impact">
<p>Process type influences the interpretation of a series and how to handle it in downstream
analytics. Behavior in a grade plant might differ drastically per regime and these changes
in behavior are not to be regarded as anomolous.
</p>
</div>
"""

import re

from typing import Optional

import networkx as nx
import numpy as np
import pandas as pd

from networkx.algorithms.components import connected_components
from pandas.api.types import is_string_dtype
from scipy.spatial.distance import pdist, squareform
from sklearn.cluster import KMeans
from pint import UndefinedUnitError

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import get_unit_registry
from timeseer.metadata import fields


META = {
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


def _get_min_max_or_bust(df):
    df = df.dropna()
    if len(df) < 30:
        return np.array([np.nan, np.nan])

    return np.nanquantile(df, [0.25, 0.75])


def _has_overlap(center1, center2):
    return not (min(center1) > max(center2) or min(center2) > max(center1))


def _no_overlap_in_relevant_clusters(clusters):
    if len(clusters) < 2:
        return False
    nb_overlap = 0
    for i in np.arange(0, len(clusters) - 1):
        for j in np.arange(i + 1, len(clusters)):
            if _has_overlap(clusters[i], clusters[j]):
                nb_overlap = nb_overlap + 1
    return (nb_overlap / len(clusters)) < 0.5


def _is_regime_cluster_based(df):
    min_resampled = df.resample("1W").apply(np.nanmin).interpolate("linear")
    max_resampled = df.resample("1W").apply(np.nanmax).interpolate("linear")
    resampled = pd.concat([min_resampled, max_resampled], axis=1).dropna().values
    if len(resampled) >= 10:
        n_clusters = 5
        kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(resampled)
        if max(kmeans.labels_) < 3:
            return False
        counts = np.bincount(kmeans.labels_)
        if max(kmeans.labels_) < (n_clusters - 1):
            counts = np.concatenate(
                (counts, [0] * (n_clusters - max(kmeans.labels_) - 1))
            )
        rel_clusters = kmeans.cluster_centers_[counts >= 3]
        return _no_overlap_in_relevant_clusters(rel_clusters)
    return False


def _is_regime_graph_based(df):
    resampled = df.resample("1W").apply(_get_min_max_or_bust).apply(pd.Series)
    dists = pdist(resampled, _has_overlap)
    square_distance_matrix = squareform(dists)

    graph = nx.from_numpy_matrix(square_distance_matrix, create_using=nx.MultiGraph)
    connected_graph_components = list(connected_components(graph))

    return (
        len(connected_graph_components) > 3
        and len([g for g in connected_graph_components if len(g) >= 3]) > 3
    )


def _is_regime(df):
    return _is_regime_cluster_based(df) or _is_regime_graph_based(df)


def _frequent_consistent_pct90(df):
    q1, q9 = np.nanquantile(df["value"], [0.1, 0.9])
    resampled = df.resample("1D").apply(lambda x: any(x > q9) and any(x < q1))
    if (sum(resampled["value"]) / len(resampled)) > 0.8:
        return True
    return False


def _is_batch(df):
    if _frequent_consistent_pct90(df):
        return True
    return False


def _get_dimensionality(ureg, unit):
    try:
        return ureg.get_dimensionality(unit)
    except (UndefinedUnitError, AttributeError, ValueError, KeyError):
        return "Unknown"


def _is_counter(analysis_input: timeseer.AnalysisInput) -> bool:
    unit = analysis_input.metadata.get_field(fields.Unit)
    if unit is not None:
        ureg = get_unit_registry()
        dimensionality = _get_dimensionality(ureg, unit)
        if dimensionality.__str__() == "[time]":
            return True
    return False


def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _check_process_type(analysis_input: timeseer.AnalysisInput) -> str:
    df = _clean_dataframe(analysis_input.data)
    if _is_counter(analysis_input):
        return "COUNTER"
    if _is_batch(df):
        return "BATCH"
    if _is_regime(df):
        return "REGIME"
    return "CONTINUOUS"


def _check_process_type_string(df: pd.DataFrame) -> str:
    if any(df["value"].str.contains("batch", flags=re.IGNORECASE)) or any(
        df["value"].str.contains("phase", flags=re.IGNORECASE)
    ):
        return "BATCH"
    if (
        any(df["value"].str.contains("regime", flags=re.IGNORECASE))
        or any(df["value"].str.contains("product", flags=re.IGNORECASE))
        or any(df["value"].str.contains("grade", flags=re.IGNORECASE))
    ):
        return "REGIME"
    return "CONTINUOUS"


def _check_process_type_dictionary(analysis_input: timeseer.AnalysisInput) -> str:
    dictionary = analysis_input.metadata.get_field(fields.Dictionary)
    if dictionary is None:
        return "REGIME"
    if any(
        True for v in dictionary.mapping.values() if "batch".upper() in v.upper()
    ) or any(True for v in dictionary.mapping.values() if "phase".upper() in v.upper()):
        return "BATCH"
    return "REGIME"


# This duplicates code in deduce_data_type check
# pylint: disable=too-many-return-statements
def _deduce_data_type(
    analysis_input: timeseer.AnalysisInput,
) -> Optional[timeseer.DataType]:
    df = analysis_input.data

    if len(_clean_dataframe(analysis_input.data)) == 0:
        return None

    if is_string_dtype(df["value"]):
        return DataType.STRING

    data_type = analysis_input.metadata.get_field(fields.DataType)
    if data_type is not None:
        if data_type == DataType.STRING and not is_string_dtype(df["value"]):
            return None
        return data_type

    if (
        df["value"].nunique() <= 10 and all(isinstance(x, int) for x in df["value"])
    ) or all(x in range(101) for x in df["value"].to_numpy()):
        return DataType.CATEGORICAL
    if isinstance(df["value"][0], np.float64):
        return DataType.FLOAT64
    if isinstance(df["value"][0], np.float32):
        return DataType.FLOAT32
    if isinstance(df["value"][0], int):
        return DataType.FLOAT32
    return None


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    result = None

    data_type = _deduce_data_type(analysis_input)
    if data_type in [DataType.FLOAT32, DataType.FLOAT64]:
        result = _check_process_type(analysis_input)
    if data_type == DataType.STRING:
        result = _check_process_type_string(analysis_input.data)
    if data_type in [DataType.DICTIONARY, DataType.CATEGORICAL]:
        result = _check_process_type_dictionary(analysis_input)

    if result is None:
        return timeseer.AnalysisResult()

    calculated_metadata = [timeseer.CalculatedMetadata("process type", result)]

    return timeseer.AnalysisResult(calculated_metadata=calculated_metadata)



"""Lower limit set in the metadata should not be subceeded.

<p>This check identifies whether the lower limit, if defined, is crossed
within the analyzed time frame.</p>
<p><img src='../static/images/reporting/limits_lower.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where the corresponding value is lower than the given limit.
Imagine that 100 points are analyzed in a given time-frame
and there are 10 points whose value is lower than the given limit.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that 90% of all points lie above the limit.</p>
<div class="ts-check-impact">
<p>
When the sensor spec limits are subceeded this is an indication of sensor failure. This might mean the
sensor needs to be recalibrated.
</p>
</div>
"""

from typing import List

import pandas as pd

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
    calculate_local_score,
)
from timeseer.metadata import fields

_CHECK_NAME = "Out-of-bounds (lower)"
_EVENT_FRAME_NAME = "Out of bounds (lower)"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame, metadata: timeseer.Metadata):
    return df["value"] < metadata.get_field(fields.LimitLowFunctional)


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _run_limit_check(analysis_input, median_archival_step, open_event_frames):
    metadata = analysis_input.metadata
    df = analysis_input.data
    df = _clean_dataframe(df)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    active_points = _get_active_points(df, metadata)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = _handle_open_intervals(df, intervals)

    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    percentage = calculate_local_score(
        df, analysis_start, intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-1]

    return percentage, list(frames), last_analyzed_point


def _is_valid_input(
    analysis_input: timeseer.AnalysisInput, median_archival_step: List[float]
):
    if analysis_input.metadata.get_field(fields.LimitLowFunctional) is None:
        return False
    if median_archival_step is None or len(median_archival_step) == 0:
        return False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return False
    return True


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: timeseer.AnalysisInput,
) -> List[timeseer.EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    if not _is_valid_input(analysis_input, median_archival_step):
        return timeseer.AnalysisResult()

    open_event_frames = _get_open_event_frames(analysis_input)

    pct_active, frames, last_analyzed_point = _run_limit_check(
        analysis_input, median_archival_step[0], open_event_frames
    )
    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult(
        [check], frames, last_analyzed_point=last_analyzed_point
    )



"""Regime detection event frames.

<p></p>"""

from datetime import timedelta
from typing import Optional

import numpy as np
import pandas as pd
import pymannkendall as mk

from pandas.api.types import is_string_dtype
from ruptures import Pelt

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe

META = {
    "event_frames": ["Regime"],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _is_frame_long_enough(frame):
    return (frame.end_date - frame.start_date) >= timedelta(weeks=1)


def _filter_event_frames(all_frames):
    filter_iterator = filter(_is_frame_long_enough, all_frames)
    return filter_iterator


def _get_event_frames(df: pd.DataFrame, breakpoints, idx, interval_type):
    event_frames = pd.DataFrame(columns=["start_date", "end_date", "type"])

    start_dates = df.iloc[breakpoints[idx] - 1]
    end_dates = df.iloc[breakpoints[idx + 1] - 1]

    event_frames["start_date"] = start_dates["value"].index
    event_frames["end_date"] = end_dates["value"].index
    event_frames["type"] = interval_type

    return event_frames_from_dataframe(event_frames)


def _run_regime_detection(df2: pd.DataFrame) -> Optional[list[timeseer.EventFrame]]:
    df = df2.resample("1D").median().interpolate("time").dropna()

    if len(df) == 0:
        return None

    values = df["value"].values.flatten()
    penalty_value = 1
    model = "l1"
    min_size = 10
    jump = 5
    algo = Pelt(model=model, min_size=min_size, jump=jump).fit(values)
    breakpoints = np.array(algo.predict(pen=penalty_value))
    #  For each section check drift with MK
    idx = [
        i
        for i in range(len(breakpoints) - 1)
        if (breakpoints[i + 1] - breakpoints[i]) > 20
        and mk.original_test(
            values[breakpoints[i] + jump : breakpoints[i + 1] - jump], alpha=0.001
        ).trend
        == "no trend"
    ]
    if len(idx) == 0 or len(breakpoints) == 0:
        return []
    return list(
        _filter_event_frames(
            _get_event_frames(df, breakpoints, np.array(idx), "Regime")
        )
    )


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if is_string_dtype(analysis_input.data["value"]):
        return timeseer.AnalysisResult()
    frames = _run_regime_detection(analysis_input.data)
    if frames is None:
        return timeseer.AnalysisResult()
    return timeseer.AnalysisResult(event_frames=frames)



"""There should be no NaN values in the data.

<p>It is possible for values to be recorded in a non-numerical format due to preprocessing or network issues.
These values are indicated as NaN (Not-a-Number).
Non-numerical reading of a sensor or dividing by zero in preprocessing are typical examples of this issue.</p>
<p><img src='../static/images/reporting/nan.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where the corresponding value is not defined (i.e. NaN).
Imagine that 100 points are analyzed in a given time-frame
and there are 10 points whose value is NaN.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that 90% of all points have valid values.</p>
<div class="ts-check-impact">
<p>
NaNs can hamper (automated) downstream analytics.
</p>
</div>
"""

from datetime import timedelta

import pandas as pd

import timeseer

from timeseer.analysis.utils import event_frames_from_dataframe

_CHECK_NAME = "NaNs"

META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Completeness",
            "event_frames": ["NaN"],
        }
    ],
    "conditions": [{"min_series": 1, "min_data_points": 1}],
    "signature": "univariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame):
    return df["value"].isna()


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_nan_check(df: pd.DataFrame, median_archival_step):
    df = _clean_dataframe(df)

    active_points = _get_active_points(df)
    intervals = _get_intervals(active_points, df, "NaN")
    check_active_time = _calculate_total_check_active_time(
        intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(intervals)

    total_time = df.index[-1] - df.index[0]
    percentage = check_active_time / total_time

    return percentage, list(frames)


def _is_valid_input(median_archival_step):
    if median_archival_step is None or len(median_archival_step) == 0:
        return False
    return True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]
    if not _is_valid_input(median_archival_step):
        return timeseer.AnalysisResult()

    pct_active, frames = _run_nan_check(analysis_input.data, median_archival_step[0])
    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult([check], frames)



"""The compression ratio based on sampling ratio."""

from timeseer.metadata import fields
from timeseer import AnalysisInput, AnalysisResult, Statistic

_STATISTIC_NAME = "Compression ratio"

META = {
    "statistics": [{"name": _STATISTIC_NAME}],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_data_points": 1,
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    sampling_rate = analysis_input.metadata.get_field(fields.SamplingRate)
    if sampling_rate is None:
        return AnalysisResult()
    meas_times = analysis_input.data.index.to_series()
    diff = meas_times[-1] - meas_times[0]
    range_seconds = diff.total_seconds()
    if range_seconds < sampling_rate:
        return AnalysisResult(statistics=[Statistic(_STATISTIC_NAME, "pct", 0)])
    original_count = int(range_seconds / sampling_rate) + 1
    compressed_count = len(analysis_input.data)
    compression_ratio = 100 * (1 - compressed_count / original_count)

    return AnalysisResult(
        statistics=[Statistic(_STATISTIC_NAME, "pct", compression_ratio)]
    )



"""Identification of sensor difference outliers between series.

<p>This check identifies periods where for redundant sensors the typical
difference is significantly altered compared to normal.</p>
<p><img src='../static/images/reporting/sensor_outlier.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where sensor drift is identified. Imagine that 100 points are analyzed in a given time-frame
and that drift is detected for 10 (consecutive) points. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no drift occurs.</p>
<div class="ts-check-impact">
<p>Changes in a physical relation between a set of series could indicate process or instrumentation issues.</p>
</div>
"""

from datetime import timedelta

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype
from scipy.stats import zscore

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Sensor profile outlier"
_EVENT_FRAME_NAME = "Sensor profile outlier"
_MIN_SERIES = 2

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Accuracy",
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": _MIN_SERIES,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        },
    ],
    "signature": "multivariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _cusum(series, threshold, anomaly_threshold):
    series = zscore(series, nan_policy="omit")
    series[np.isnan(series)] = 0
    positive_cusum = (series * 0).copy()
    negative_cusum = positive_cusum.copy()
    for i in np.arange(series.shape[0]):
        if i == 0:
            positive_cusum[i, :] = series[i, :]
            negative_cusum[i, :] = series[i, :]
        else:
            positive_cusum[i, :] = np.maximum(
                0, series[i, :] - threshold + positive_cusum[i - 1, :]
            )
            negative_cusum[i, :] = np.maximum(
                0, -threshold - series[i, :] + negative_cusum[i - 1, :]
            )
    anomalies = (positive_cusum > anomaly_threshold) | (
        negative_cusum > anomaly_threshold
    )
    return np.any(anomalies, axis=1)


def _get_intervals(df, anomalies):
    anomalies = pd.Series(data=anomalies, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME

    return intervals


def _make_diff_array(df):
    diff_array = np.zeros((df.shape[0], int(df.shape[1] * (df.shape[1] - 1) / 2)))
    idx = 0
    for i in range(df.shape[1] - 1):
        for j in range(i + 1, df.shape[1]):
            diff_array[:, idx] = df.iloc[:, i] - df.iloc[:, j]
            idx = idx + 1
    return diff_array


def _profile_outlier(concatenated_df):
    diff_array = _make_diff_array(concatenated_df)
    return _cusum(diff_array, 0.5, 3)


def _get_mv_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _clean_input(inputs, selected_index):
    index = inputs[selected_index].data.index
    return (
        pd.concat(
            [
                series.data["value"][
                    ~series.data["value"].index.duplicated(keep="first")
                ]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .reindex(index)
        .interpolate("linear")
        .dropna()
        .sort_index()
    )


def _get_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _run_mv_redundant_sensor_outlier(inputs: list[timeseer.AnalysisInput]):
    selected_index = np.argmax(
        [_get_median_archival_step(inputs[i].data) for i in range(len(inputs))]
    )
    concatenated_df = _clean_input(inputs, selected_index)
    if len(concatenated_df) < 30:
        return None, []

    median_archival_step = _get_mv_median_archival_step(concatenated_df)

    anomalies = _profile_outlier(concatenated_df)
    intervals = _get_intervals(concatenated_df, anomalies)
    check_active_time = _calculate_total_check_active_time(
        intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(intervals)

    total_time = concatenated_df.index[-1] - concatenated_df.index[0]
    percentage = check_active_time / total_time

    return percentage, list(frames)


def _filter_invalid_inputs(
    inputs: list[timeseer.AnalysisInput],
) -> list[timeseer.AnalysisInput]:
    valid_inputs = []
    for check_input in inputs:
        if is_string_dtype(check_input.data["value"]):
            continue
        valid_inputs.append(check_input)
    return valid_inputs


def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = _filter_invalid_inputs(analysis_input.inputs)
    if len(inputs) < _MIN_SERIES:
        return timeseer.AnalysisResult()

    pct_active, frames = _run_mv_redundant_sensor_outlier(inputs)
    if pct_active is None:
        return timeseer.AnalysisResult()

    check = timeseer.CheckResult(_CHECK_NAME, float(pct_active))
    return timeseer.AnalysisResult(check_results=[check], event_frames=frames)



""" Calculate the distributed quantile sketch of the time series. """

import jsonpickle
import numpy as np

from ddsketch.ddsketch import DDSketch
from pandas.api.types import is_string_dtype

import timeseer

from timeseer import DataType


META: dict = {
    "statistics": [
        {"name": "Archival Sketch"},
        {"name": "Extremes Archival Sketch"},
        {"name": "Base Archival Sketch"},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64, None],
        }
    ],
    "signature": "univariate",
}


def _get_up_quantiles_and_iqr_from_values(values):
    rel_values = values[values > 0]
    if len(rel_values) == 0:
        return None, None, None
    q25, q75 = np.nanquantile(rel_values, [0.25, 0.75])
    iqr = q75 - q25
    return q25, q75, iqr


def _get_down_quantiles_and_iqr_from_values(values):
    rel_values = values[values < 0]
    if len(rel_values) == 0:
        return None, None, None
    q25, q75 = np.nanquantile(rel_values, [0.25, 0.75])
    iqr = q75 - q25
    return q25, q75, iqr


def _get_quantiles_and_iqr_from_sketch(sketch):
    q25, q75 = [sketch.get_quantile_value(q) for q in [0.25, 0.75]]
    iqr = q75 - q25
    return q25, q75, iqr


def _get_diff_times(df):
    meas_times = df.index.to_series()
    return (meas_times - meas_times.shift()).dt.total_seconds()


def _get_extremes_archival_sketch(
    diff_times, diff_values, extremes_archival_sketch, jump_up_sketch, jump_down_sketch
):
    if extremes_archival_sketch is None:
        extremes_archival_sketch = DDSketch(0.001)

    if jump_up_sketch is None:
        _, up_q75, up_iqr = _get_up_quantiles_and_iqr_from_values(diff_values)
    if jump_up_sketch is not None:
        _, up_q75, up_iqr = _get_quantiles_and_iqr_from_sketch(jump_up_sketch)
    if jump_down_sketch is None:
        down_q25, _, down_iqr = _get_down_quantiles_and_iqr_from_values(diff_values)
    if jump_down_sketch is not None:
        down_q25, _, down_iqr = _get_quantiles_and_iqr_from_sketch(jump_down_sketch)

    ups = np.array([False] * len(diff_values))
    downs = np.array([False] * len(diff_values))

    if up_q75 is not None:
        ups = np.array(diff_values > (up_q75 + 3 * up_iqr))
    if down_q25 is not None:
        downs = np.array(diff_values < (down_q25 - 3 * down_iqr))

    extremes = ups | downs

    for v in diff_times[extremes]:
        extremes_archival_sketch.add(v)

    return extremes_archival_sketch


# pylint: disable=too-many-locals
def _get_base_archival_sketch(
    diff_times, diff_values, base_archival_sketch, jump_up_sketch, jump_down_sketch
):
    if base_archival_sketch is None:
        base_archival_sketch = DDSketch(0.001)

    if jump_up_sketch is None:
        up_q25, up_q75, _ = _get_up_quantiles_and_iqr_from_values(diff_values)
    if jump_up_sketch is not None:
        up_q25, up_q75, _ = _get_quantiles_and_iqr_from_sketch(jump_up_sketch)
    if jump_down_sketch is None:
        down_q25, down_q75, _ = _get_down_quantiles_and_iqr_from_values(diff_values)
    if jump_down_sketch is not None:
        down_q25, down_q75, _ = _get_quantiles_and_iqr_from_sketch(jump_down_sketch)

    ups = np.array([False] * len(diff_values))
    downs = np.array([False] * len(diff_values))

    if up_q75 is not None:
        ups = np.array(diff_values <= up_q75) & np.array(diff_values >= up_q25)
    if down_q25 is not None:
        downs = np.array(diff_values <= down_q75) & np.array(diff_values >= down_q25)

    bases = ups | downs

    for v in diff_times[bases]:
        base_archival_sketch.add(v)

    return base_archival_sketch


def _get_archival_sketch(diff_times, archival_sketch):
    if archival_sketch is None:
        archival_sketch = DDSketch(0.001)

    for v in diff_times:
        archival_sketch.add(v)

    return archival_sketch


def _clean_dataframe(df):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


# pylint: disable=too-many-arguments
def _run_get_sketch(
    analysis_input,
    archival_sketch,
    extremes_archival_sketch,
    base_archival_sketch,
    jump_up_sketch,
    jump_down_sketch,
):
    df = _clean_dataframe(analysis_input.data)

    diff_times = _get_diff_times(df)
    diff_values = df["value"] - df["value"].shift()

    archival_sketch = _get_archival_sketch(diff_times, archival_sketch)
    extremes_archival_sketch = _get_extremes_archival_sketch(
        diff_times,
        diff_values,
        extremes_archival_sketch,
        jump_up_sketch,
        jump_down_sketch,
    )
    base_archival_sketch = _get_base_archival_sketch(
        diff_times, diff_values, base_archival_sketch, jump_up_sketch, jump_down_sketch
    )

    return archival_sketch, extremes_archival_sketch, base_archival_sketch


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return jsonpickle.decode(statistics[0])


def _is_valid_input(analysis_input: timeseer.AnalysisInput):
    if is_string_dtype(analysis_input.data["value"]):
        return False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return False
    return True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if not _is_valid_input(analysis_input):
        return timeseer.AnalysisResult()

    archival_sketch = _get_relevant_statistic(analysis_input, "Archival Sketch")
    extremes_archival_sketch = _get_relevant_statistic(
        analysis_input, "Extremes Archival Sketch"
    )
    base_archival_sketch = _get_relevant_statistic(
        analysis_input, "Base Archival Sketch"
    )
    jump_up_sketch = _get_relevant_statistic(analysis_input, "Jump Up Sketch")
    jump_down_sketch = _get_relevant_statistic(analysis_input, "Jump Down Sketch")

    archival_sketch, extremes_archival_sketch, base_archival_sketch = _run_get_sketch(
        analysis_input,
        archival_sketch,
        extremes_archival_sketch,
        base_archival_sketch,
        jump_up_sketch,
        jump_down_sketch,
    )

    statistics = []
    if archival_sketch is not None:
        statistics.append(
            timeseer.Statistic(
                META["statistics"][0]["name"],
                "sketch",
                jsonpickle.encode(archival_sketch),
            )
        )
    if extremes_archival_sketch is not None:
        statistics.append(
            timeseer.Statistic(
                META["statistics"][1]["name"],
                "sketch",
                jsonpickle.encode(extremes_archival_sketch),
            )
        )
    if base_archival_sketch is not None:
        statistics.append(
            timeseer.Statistic(
                META["statistics"][2]["name"],
                "sketch",
                jsonpickle.encode(base_archival_sketch),
            )
        )

    if len(statistics) == 0:
        return timeseer.AnalysisResult()

    return timeseer.AnalysisResult(statistics=statistics)



"""Recommendation for low and high limits based on historical data.

<p>Each sensor or lab measuring equipment has an inherent measurement range. Outside of this range
the readings from the sensor can't be relied on. This information is typically present in
the sensor specification sheet.
Traditional historians also allow setting limits (often Low-Low, Low, High and High-High limits).
These limits are often based on process safety concerns or indeed taken from the normal operating range.
In all cases having access to these limits is essential for interpreting readings from the equipment.
The <b>recommended limits</> are based upon historical measurements and rounded to a human friendly number
similar as in most charting.</p>
<div class="ts-check-impact">
<p>Missing or badly configured limits can cause missing warning signs of abnormal situations both
in the process as well as with the instrumentation.
</p>
</div>
"""

import pandas as pd
import numpy as np

from pandas.api.types import is_string_dtype

from pint import UndefinedUnitError

import timeseer

from timeseer import DataType
from timeseer.analysis.utils import get_unit_registry
from timeseer.metadata import fields


META = {
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 12,
            "min_data_points": 300,
        }
    ],
    "signature": "univariate",
}


def _get_dimensionality(unit):
    ureg = get_unit_registry()
    try:
        return ureg.get_dimensionality(unit)
    except (UndefinedUnitError, AttributeError, ValueError, KeyError):
        return "Unknown"


def _make_units_dir():
    ureg = get_unit_registry()
    units_zero_min = [
        "meter",
        "second",
        "ampere",
        "candela",
        "gram",
        "mole",
        "kelvin",
        "unit",
        "pH",
        "m2",
        "liter",
        "hertz",
        "kph",
        "galileo",
        "newton",
        "joule",
        "watt",
        "water",
        "pascal",
        "foot_pound",
        "poise",
        "stokes",
        "rhe",
        "particle",
        "molar",
        "katal",
        "clausius",
        "entropy_unit",
        "curie",
        "langley",
        "nit",
        "lumen",
        "lux",
        "a_u_intensity",
        "volt",
        "ohm",
        "siemens",
        "henry",
        "weber",
        "tesla",
        "bohr_magneton",
    ]
    dimensions_zero_min = []
    for unit in units_zero_min:
        dimension = ureg.get_dimensionality(unit)
        dimensions_zero_min.append(dimension)
    return dimensions_zero_min


def _unit_dir(analysis_input: timeseer.AnalysisInput):
    input_dimension = _get_dimensionality(
        analysis_input.metadata.get_field(fields.Unit)
    )
    return input_dimension


def _deconstruct(value):
    if value == 0:
        return [0, 0, 0, 0]
    abs_value = abs(value)
    sign = np.sign(value)
    magnitude = np.floor(np.log10(abs_value))
    scale = np.floor(abs_value // np.power(10, magnitude))
    remainder = abs_value - scale * np.power(10, magnitude)
    smsr = [sign, magnitude, scale, remainder]
    return smsr


def _max_limit(max_value):
    smsr = _deconstruct(max_value)
    sign = smsr[0]
    magnitude = smsr[1]
    scale = smsr[2]
    remainder = smsr[3]
    if magnitude < 0:
        if sign < 0:
            return 0
        lim = np.power(10, magnitude + 1)
        return lim
    lim = np.power(10, magnitude) * (scale)
    if remainder > 0:
        if sign > 0:
            smsr_remainder = _deconstruct(remainder)
            small_mag = max((smsr_remainder[1] + 1), np.floor(magnitude / 2))
            small = np.power(10, np.sign(magnitude) * small_mag)
            lim += small
        if sign < 0:
            smsr_remainder = _deconstruct(remainder)
            if smsr_remainder[1] >= (magnitude / 2):
                small = np.power(10, smsr_remainder[1]) * smsr_remainder[2]
                lim += small
    return float(sign * lim)


def _min_limit(min_value):
    smsr = _deconstruct(min_value)
    sign = smsr[0]
    magnitude = smsr[1]
    scale = smsr[2]
    remainder = smsr[3]
    lim = 0
    if sign < 0:
        if magnitude < 0:
            return -1
        if magnitude >= 0:
            lim = np.power(10, magnitude) * (scale)
            if remainder > 0:
                smsr_remainder = _deconstruct(remainder)
                small_mag = max((smsr_remainder[1] + 1), np.floor(magnitude / 2))
                small = np.power(10, np.sign(magnitude) * small_mag)
                lim += small
    return float(sign * lim)


def _find_recommended_upper_limit(df: pd.DataFrame):
    if len(df["value"]) < 100:
        max_value = max(df["value"])
    else:
        q_1, q_3 = np.quantile(df["value"], [0.25, 0.75])
        iqr = q_3 - q_1
        max_value = max(df["value"][df["value"] <= q_3 + 1.5 * iqr])

    rec_high = 0
    if max_value != 0:
        rec_high = _max_limit(max_value)
    return rec_high


def _find_recommended_lower_limit(df: pd.DataFrame):
    if len(df["value"]) < 100:
        min_value = min(df["value"])
    else:
        q_1, q_3 = np.quantile(df["value"], [0.25, 0.75])
        iqr = q_3 - q_1
        min_value = min(df["value"][df["value"] >= q_1 - 1.5 * iqr])

    rec_low = 0
    if min_value != 0:
        rec_low = _min_limit(min_value)
    return rec_low


def _is_valid_input(analysis_input: timeseer.AnalysisInput) -> bool:
    if is_string_dtype(analysis_input.data["value"]):
        return False
    data_type = analysis_input.metadata.get_field(fields.DataType)
    if data_type not in [DataType.FLOAT64, DataType.FLOAT32, None]:
        return False
    return True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    if not _is_valid_input(analysis_input):
        return timeseer.AnalysisResult()

    data = analysis_input.data.dropna()
    if len(data) == 0:
        return timeseer.AnalysisResult()

    dimensions_zero_min = _make_units_dir()

    if _unit_dir(analysis_input) in dimensions_zero_min:
        recommended_lower_limit = 0
    else:
        recommended_lower_limit = _find_recommended_lower_limit(data)

    if _unit_dir(analysis_input) == _get_dimensionality("pH"):
        recommended_upper_limit = 14
    else:
        recommended_upper_limit = _find_recommended_upper_limit(data)

    calculated_metadata = [
        timeseer.CalculatedMetadata("functional lower limit", recommended_lower_limit),
        timeseer.CalculatedMetadata("functional upper limit", recommended_upper_limit),
    ]
    return timeseer.AnalysisResult(calculated_metadata=calculated_metadata)



"""The description should not match the series name.

<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check belongs to the score group Description.
If this check is True, it means that the score for group Description will be 0%.</p>
<div class="ts-check-impact">
<p>An interpretable description of series aids in onboarding and analysis by 3rd
parties that are less familiar with the process.
Due to time constraints sometimes the non-descriptive name of the series is copied for the description.
A good description provides information on the type as well as location of the sensor.
</p>
</div>
"""

import timeseer

from timeseer.metadata import fields

_CHECK_NAME = "Description does not equal series name"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Description",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    description = analysis_input.metadata.get_field(fields.Description)
    if description == "":
        return timeseer.AnalysisResult()

    if description == analysis_input.metadata.series.name:
        score = 1.0
    else:
        score = 0.0
    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, score)]
    )



"""The series should have sensical values for the limits.

<p>Each sensor or lab measuring equipment has an inherent measurement range. Outside of this range
the readings from the sensor are unreliable. This information is typically present in the
sensor specification sheet.
Traditional historians also allow setting limits (often Low-Low, Low, High and High-High limits).
These limits are often based on process safety concerns or indeed taken from the normal operating range.
In all cases having access to these limits is essential for interpreting readings from the equipment.
Because of time constraints or as a placeholder it can occur that for all limits a zero value is filled in.
It is also possible that due to simple oversight the limits are switched and for instance the higher limit
is less than the lower limit</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check belongs to the score group Limits.
If this check is True, it means that the score for group Limits will be 0%.
</p>
<div class="ts-check-impact">
<p>Missing or badly configured limits can cause missing warning signs of abnormal situations both
in the process as well as with the instrumentation.
</p>
</div>
"""

import timeseer

from timeseer.metadata import fields

_CHECK_NAME = "Sensical physical limits"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Limits",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(analysis_input: timeseer.AnalysisInput) -> timeseer.AnalysisResult:
    limit_low = analysis_input.metadata.get_field(fields.LimitLowPhysical)
    limit_high = analysis_input.metadata.get_field(fields.LimitHighPhysical)
    if limit_low is None or limit_high is None:
        return timeseer.AnalysisResult()
    if limit_low >= limit_high:
        score = 1.0
    else:
        score = 0.0
    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, score)]
    )



"""For percentage accuracy the value should be between 0 and 100.

<p>This check validates whether the accuracy percentage falls
within the expected range between 0 and 100.</p>
<p class="scoring-explanation">The score for this checks is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Badly configured accuracy percentage can cause a wrong accuracy calculation.
</p>
</div>
"""

import timeseer

from timeseer import DataType
from timeseer.metadata import fields

_CHECK_NAME = "Accuracy percentage is within range"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Metadata",
            "group": "Accuracy",
            "data_type": "bool",
        }
    ],
    "conditions": [
        {"min_series": 1, "data_type": [DataType.FLOAT32, DataType.FLOAT64]}
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    accuracy_percentage = analysis_input.metadata.get_field(fields.AccuracyPercentage)
    if accuracy_percentage is None:
        return timeseer.AnalysisResult()
    if accuracy_percentage < 0 or accuracy_percentage > 100:
        score = 1.0
    else:
        score = 0.0
    return timeseer.AnalysisResult(
        check_results=[timeseer.CheckResult(_CHECK_NAME, score)]
    )
